use crate::felinet::handler_client::HandlerClient;

use anyhow::{anyhow, bail, Context};
use clap::{arg, Parser};
use config::Config;
use felinet::PingRequest;
use gate::{beacon_forever, felinet_receive_forever, felinet_send_forever, packet_handler};
use ham_cats::{
    packet::Packet,
    whisker::{Arbitrary, Identification},
};
use radio::{RadioManager, MAX_PACKET_LEN};
use std::{str::FromStr, time::Duration};
use tokio::sync::{broadcast, mpsc};
use tonic::{transport::Endpoint, Request};
use uuid::Uuid;

mod config;
mod gate;
mod radio;
mod util;

pub mod felinet {
    tonic::include_proto!("felinet");
}

#[derive(Parser)]
struct Cli {
    /// Repeatedly send very long packets on the specified frequency for debugging purposes.
    /// Can be used when measuring power output, for example.
    /// Valid frequencies are 430MHz - 436.375MHz, in 25kHz increments
    #[arg(long, short, name = "freq")]
    xmit_test: Option<f64>,
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::load()?;
    let cli = Cli::parse();

    let (packet_tx, packet_receive) = mpsc::channel(16);
    let (packet_send, packet_rx) = mpsc::channel(16);
    let radio = RadioManager::new(packet_tx, packet_rx).unwrap();

    match cli.xmit_test {
        Some(freq) => xmit_test(&config, freq, radio, packet_send).await?,
        None => gate_forever(&config, radio, packet_send, packet_receive).await?,
    }

    Ok(())
}

async fn xmit_test(
    config: &Config,
    freq: f64,
    mut radio: RadioManager,
    packet_send: mpsc::Sender<Vec<u8>>,
) -> anyhow::Result<()> {
    if !(430.0..=436.380).contains(&freq) {
        bail!("Frequency {} MHz out of range (430MHz - 436.375MHz)", freq);
    }

    let channel = ((freq - 430.0) / 0.025).round() as u8;

    radio.set_channel(channel);

    let mut pkt: Packet<MAX_PACKET_LEN> = Packet::default();
    pkt.add_identification(
        Identification::new(config.icon, &config.callsign, config.ssid)
            .context("Invalid identification")?,
    )
    .map_err(|e| anyhow!("Could not add identification to packet: {e}"))?;
    pkt.add_comment("Debugging packet")
        .map_err(|e| anyhow!("Could not add comment to packet: {e}"))?;

    for _ in 0..((MAX_PACKET_LEN - 100) / 600) {
        pkt.add_arbitrary(Arbitrary::new(&[0xA5; 255]).unwrap())
            .map_err(|e| anyhow!("Could not add data to packet: {e}"))?;
    }

    let data = pkt
        .fully_encode()
        .map_err(|e| anyhow!("Could not encode packet: {e}"))?;

    tokio::task::spawn(async move {
        loop {
            if let Err(e) = radio.process_forever().await {
                eprintln!("Radio error: {e}")
            }
        }
    });

    let actual_freq = 430.0 + 0.025 * channel as f64;
    println!("Transmitting test data on {actual_freq:.3} MHz... Press Ctrl+C to stop");

    loop {
        packet_send.send(data.to_vec()).await?;
    }
}

async fn gate_forever(
    config: &Config,
    mut radio: RadioManager,
    packet_send: mpsc::Sender<Vec<u8>>,
    packet_receive: mpsc::Receiver<Vec<u8>>,
) -> anyhow::Result<()> {
    let uuid = Uuid::new_v4();

    tokio::task::spawn(async move {
        loop {
            if let Err(e) = radio.process_forever().await {
                eprintln!("Radio error: {e}")
            }
        }
    });

    let tx = broadcast::Sender::new(16);

    beacon_forever(config, packet_send.clone(), tx.clone(), uuid)?;

    let endpoint = Endpoint::from_str(&config.felinet.address)?
        .keep_alive_while_idle(true)
        .keep_alive_timeout(Duration::from_secs(5))
        .http2_keep_alive_interval(Duration::from_secs(5))
        .tcp_keepalive(Some(Duration::from_secs(5)));
    let client = HandlerClient::connect(endpoint).await?;
    let callsign = config.callsign.clone();

    {
        let tx = tx.clone();
        let packet_send = packet_send.clone();
        let ssid = config.ssid;
        let callsign = config.callsign.clone();

        tokio::task::spawn(async move {
            if let Err(e) = packet_handler(
                callsign.clone(),
                ssid,
                tx,
                packet_send,
                packet_receive,
                uuid,
            )
            .await
            {
                eprintln!("Packet handler error: {}. Aborting", e)
            }
        });
    }

    {
        let client = client.clone();
        tokio::task::spawn(async {
            felinet_send_forever(client, tx).await;
        });
    }

    {
        let mut client = client.clone();
        tokio::task::spawn(async move {
            loop {
                client
                    .ping(Request::new(PingRequest { uuid: uuid.into() }))
                    .await
                    .ok();

                tokio::time::sleep(Duration::from_secs(600)).await;
            }
        });
    }

    felinet_receive_forever(client, packet_send, &callsign, config.ssid, uuid).await;

    Ok(())
}
